from typing import Any

from haco import trigger
from haco.sensor import Sensor
from haco.tasmota import Tasmota


def get_memory_sensor():
    memory_sensor = Sensor('Memory Free', uom='kB', device_class='DATA_SIZE', data_type=int)

    @memory_sensor.callback(trigger.Cron('*/30 * * * * *'))
    def tasmota(value: Tasmota[Any, 'tasmota.memory()']):
        return value.get('heap_free', 'unknown')

    return memory_sensor



from haco.climate import Climate
from haco.number import Number
from haco.pulldown import Select
from haco.sensor import Sensor
from haco.tasmota import Tasmota
from haco.text import Text, Password

number = Number(
    'Dev Number Test Control',
    number_range=range(3, 91),
    step=0.5,
    mode='box',
    uom='mm'
)


@number.callback()
def ha(value: str) -> Tasmota[float, 'tasmota.cmd("var10 "+str(value))']:
    return value


@number.callback('var10#state')
def tasmota(value: float, control):
    return value


climate = Climate(
    'Dev Climate Test Control',
    swing_modes=["Fixed 90°", "Swing", "Fixed 45°"],
    preset_modes=['Smart', 'Purify', 'High', 'Low'],
    modes=['dry', 'fan_only', 'off'],
    temperature_range=range(10, 90),
    humidity_range=range(20, 80)

)


@climate.callback()
def mode_ha(value: str) -> Tasmota[int, 'tasmota.cmd("var1 "+str(value))']:
    return ['dry', 'fan_only', 'off'].index(value)


@climate.callback('var1#state')
def mode_tasmota(value: int):
    return ['dry', 'fan_only', 'off'][value]


@climate.callback()
def target_humidity_ha(value: int) -> Tasmota[int, 'tasmota.cmd("var2 "+str(value))']:
    return value


@climate.callback('var2#state')
def target_humidity_tasmota(value: int):
    return value


@climate.callback('var2#state')
def current_humidity_tasmota(value: int):
    return value + 5


@climate.callback('var3#state')
def current_temperature_tasmota(value: int):
    return value + 10


@climate.callback()
def target_temperature_ha(value: float) -> Tasmota[float, 'tasmota.cmd("var3 "+str(value))']:
    return value


@climate.callback('var3#state')
def target_temperature_tasmota(value: int):
    return value


@climate.callback('var1#state')
def action_tasmota(value: int, is_on: Tasmota[bool, 'tasmota.get_power()[0]']):
    if not is_on:
        return 'off'

    return 'fan' if value == 1 else 'drying'


@climate.callback()
def swing_mode_ha(value: str) -> Tasmota[int, 'tasmota.cmd("var4 "+str(value))']:
    return ['Fixed 90°', 'Fixed 45°', 'Swing'].index(value)


@climate.callback('var4#state')
def swing_mode_tasmota(value: int):
    return ['Fixed 90°', 'Fixed 45°', 'Swing'][value]


@climate.callback()
def preset_mode_ha(value: str) -> Tasmota[int, 'tasmota.cmd("var7 "+str(value))']:
    return ['Smart', 'Purify', 'High', 'Low'].index(value)


@climate.callback('var7#state')
def preset_mode_tasmota(value: int):
    return ['Smart', 'Purify', 'High', 'Low'][value]


@climate.callback()
def temperature_high_ha(value: float) -> Tasmota[float, 'tasmota.cmd("var5 "+str(value))']:
    return value


@climate.callback('var5#state')
def temperature_high_tasmota(value: int):
    return value


@climate.callback()
def temperature_low_ha(value: float) -> Tasmota[float, 'tasmota.cmd("var6 "+str(value))']:
    return value


@climate.callback('var6#state')
def temperature_low_tasmota(value: int):
    return value


pulldown = Select(
    'Dev Select Test Control',
    options=list('ABC')
)


@pulldown.callback()
def ha(value: str) -> Tasmota[str, 'tasmota.cmd("var12 "+str(value))']:
    return value


@pulldown.callback('var12#state')
def tasmota(value: str):
    return value


sensor = Sensor('Dev Sensor Test Control')


@sensor.callback('var13#state')
def tasmota(value: str):
    return value


text = Text('Dev Text Test Control')
password = Password('Dev Password Test Control')


@text.callback()
def ha(value: str) -> Tasmota[str, 'tasmota.cmd("var11 "+str(value))']:
    return value


@text.callback('var11#state')
def tasmota(value: str):
    return value


@password.callback()
def ha(value: str) -> Tasmota[str, 'tasmota.cmd("var14 "+str(value))']:
    return value


@password.callback('var14#state')
def tasmota(value: str):
    return value



from haco import schema
from haco.control import Control


class Button(Control):
    NAME = 'button'

    @classmethod
    def get_schema(cls):
        return schema.Schema(
            capabilities=[
                schema.Capability(
                    name='state',
                    ha=schema.HomeAssistant(announce_data=schema.AnnounceTopic(key='{io_ha}_topic')),
                    tamota=None
                )
            ]
        )

    def __init__(self, name, icon=None):
        super().__init__(name, icon=icon)



from haco import tuya
from .button import Button
from .number import Number
from .pulldown import Select
from .sensor import Sensor
from .switch import Switch
from .tasmota import Tasmota

__all__ = [
    "Button",
    "Select",
    "Sensor",
    "Tasmota",
    "Number",
    "Switch",
    "tuya"
]



from haco.control import Control

from haco.tools import get_range_pair


class Number(Control):
    NAME = 'number'

    # def init(name, number_range, mode, step, uom, entity_id, icon, callbacks)
    def __init__(self, name, number_range: range, icon=None, mode=None, step=None, uom=None):

        self.min, self.max = get_range_pair(number_range)
        self.mode = mode
        self.step = step
        self.uom = uom

        super().__init__(name, icon=icon)

    @property
    def control_type(self):

        if self.step is None or self.step == 1:
            return int
        else:
            return float

    def get_config_ha_ex(self):
        data = {
            'min': self.min,
            'max': self.max,
            'mode': self.mode,
            'unit_of_measurement': self.uom,
            'step': self.step
        }

        return data



from haco.control import Control

from haco.tools import invert_bool


class Switch(Control):
    NAME = 'switch'

    def __init__(self, name, icon=None):
        super().__init__(name, icon=icon)

    @property
    def control_type(self):
        return invert_bool



from collections import namedtuple

import inspect
import typing
from typing import _AnnotatedAlias

from haco import berry
from haco.trigger import Trigger, Rule

Signature = namedtuple('Signature', ['type', 'expression'])


class Callback:

    def __init__(self, capability, platform, trigger, function):

        self.platform = platform
        self.function = function
        self.capability = capability

        if trigger:
            if not isinstance(trigger, Trigger):
                trigger = Rule(trigger=trigger)
            self.trigger = trigger.trigger
            self.type_id = trigger.__class__.__name__
        else:
            self.trigger = trigger

        self.validate()

    def get_function_data(self):
        len_data = len(Signature._fields)
        signature = inspect.signature(self.function)
        arg_data = {param.name: Signature(*([None] * len_data)) for param in signature.parameters.values()}
        hints = typing.get_type_hints(self.function, include_extras=True)
        hints.setdefault('return', None)

        for key, value in hints.items():
            if type(value) is _AnnotatedAlias:
                value = typing.get_args(value)
            else:
                value = [value]
            value = [*value, *[None] * (len_data - len(value))]
            arg_data[key] = Signature(*value)

        return arg_data

    def is_berry_function(self, string: str):

        return (stripped := string.strip()).startswith('def') or stripped.startswith('/')

    def get_berry_exp(self, exp):

        if not self.is_berry_function(exp):
            return exp

        return f'({exp}(value,data))'

    @property
    def function_tasmota(self):

        arg_data = self.get_function_data()

        return_data = arg_data.pop('return')

        known_args = {'value', 'data'}
        skip_args = {'control'}

        if self.platform == 'tasmota':

            berry_exp = {}
            for name, (type, exp) in arg_data.items():
                if not exp:

                    if name in skip_args:
                        continue

                    if name not in known_args:
                        msg = f'Callback {self.function_name} argument {name} requires a Berry expression annotation.'
                        raise ValueError(msg)

                    exp = name
                berry_exp[name] = self.get_berry_exp(exp)

            berry_exp_str = ','.join([f"{repr(key)}:{value}" for key, value in berry_exp.items()])
            berry_exp_str = f"{{{berry_exp_str}}}"
            berry_exp_str = berry.Expression(berry_exp_str)

        else:

            if not return_data.expression:
                msg = f'Callback function for {self.function_name} return type requires a Berry expression annotation.'
                raise ValueError(msg)

            berry_exp_str = berry.Expression(self.get_berry_exp(return_data.expression))

        return str(berry_exp_str)

    @property
    def function_name(self):
        return self.function.__name__

    def validate(self):

        self.function_tasmota

        if self.platform == 'tasmota':
            if not self.trigger:
                raise ValueError(f'Callback {self.function_name} must have a trigger.')



import hashlib
import logging

from haco import constants, tuya
from haco.tasmota import Tasmota

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)




def split_into_chunks(input_string, chunk_size):
    return [input_string[i:i + chunk_size] for i in range(0, len(input_string), chunk_size)]


def get_range_pair(data: range):
    if not data:
        return None, None

    return data.start, data.stop


def truncate_middle(input_string, max_length):
    if len(input_string) <= max_length:
        return input_string

    len_start = (max_length - 3) // 2
    len_end = max_length - 3 - len_start

    truncated = f'{input_string[:len_start]}...{input_string[-len_end:]}'

    return truncated


async def log_publish(client, topic, payload, retain=False):
    log_retain = '|' if retain else ' '
    message = f'mqtt->{log_retain} {topic} {truncate_middle(payload, 100)}'
    logger.debug(message)
    await client.publish(topic=topic, payload=payload, retain=retain)


async def log_subscribe(client, topic):
    message = f'mqtt+{topic}'
    logger.debug(message)
    await client.subscribe(topic=topic)


async def log_received(message):
    log_retain = '|' if message.retain else ' '
    message = f'mqtt<-{log_retain} {message.topic} {truncate_middle(message.payload.decode("utf-8"), 100)}'
    logger.debug(message)


def callback_default(value, data=None):
    return value


def invert_bool(value):
    return {True: 'ON', False: 'OFF', 'ON': True, 'OFF': False}[value]


def sanitize_name(s, sep='-'):
    chars = []

    for c in s.lower():
        if c in constants.SEPS_ALLOWED:
            chars.append(sep)
        elif c in constants.CHARS_ALLOWED:
            chars.append(c)

    sanitized = ''.join(chars)

    if not sanitized:
        raise ValueError("Sanitized string is empty")

    return sanitized


def add_tuya_io(control, control_type, type_id, dp_id):
    @control.callback(trigger=tuya.received(type_id, dp_id))
    def tasmota(value: control_type):
        return value

    @control.callback()
    def ha(value: control_type) -> Tasmota[control_type, tuya.send(type_id, dp_id)]:
        return value


def hash_string(input_string, size=16):
    # Create a hashlib object for the SHA-256 algorithm
    hash_object = hashlib.sha256()

    # Update the hash object with the bytes of the input string
    hash_object.update(input_string.encode('utf-8'))

    # Get the hexadecimal representation of the hash
    hash_result = hash_object.hexdigest()[:size]

    return hash_result



import math

from haco import Tasmota, Number, Sensor, Select, Switch, tuya
from haco.tools import add_tuya_io

OPTIONS = ['Default', 'Fries', 'Shrimp', 'Pizza', 'Chicken', 'Fish', 'Steak', 'Cake', 'Bacon', 'Preheat', 'Custom']

cookbook = Select('Cookbook', options=OPTIONS, icon='mdi:chef-hat')


@cookbook.callback()
def ha(value: str) -> Tasmota[int, tuya.send(4, 3)]:
    return OPTIONS.index(value)


@cookbook.callback(trigger=tuya.received(4, 3))
def tasmota(value: int):
    return OPTIONS[value]


time_remaining = Sensor(
    name='Time Remaining',
    uom='minutes',
    icon='mdi:timer',
    device_class='DURATION'
)


@time_remaining.callback(trigger=tuya.received(2, 8))
def tasmota(value: int):
    return value


cooking_time = Number(
    name='Cooking Time',
    number_range=range(1, 60),
    uom='minutes',
    icon='mdi:timer'
)


@cooking_time.callback(trigger=tuya.received(2, 7))
def tasmota(value: int):
    return value


@cooking_time.callback()
def ha(value: int) -> Tasmota[int, tuya.send(2, 7)]:
    return value


cooking_temp = Number(
    name='Cooking Temperature (F)',
    number_range=range(1, 60),
    uom='minutes',
    icon='mdi:timer'
)


@cooking_temp.callback(trigger=tuya.received(2, 7))
def tasmota(value: int):
    return value


@cooking_temp.callback()
def ha(value: int) -> Tasmota[int, tuya.send(2, 7)]:
    return value


cooking_temp_f = Number(
    'Cooking Temperature (F)',
    number_range=range(170, 399),
    mode='slider',
    uom='°F',
    icon='mdi:temperature-fahrenheit'
)


@cooking_temp_f.callback()
def ha(value: float) -> Tasmota[int, tuya.send(2, 103)]:
    return value


@cooking_temp_f.callback(trigger=tuya.received(2, 103))
def tasmota(value: float):
    return value


cooking_temp_c = Number(
    'Cooking Temperature (C)',
    number_range=range(77, 204),
    mode='slider',
    uom='°C',
    icon='mdi:temperature-celsius'
)


@cooking_temp_c.callback()
def ha(value: float) -> Tasmota[int, tuya.send(2, 103)]:
    return (value * 1.8) + 32


@cooking_temp_c.callback(trigger=tuya.received(2, 103))
def tasmota(value: float):
    return math.ceil((value - 32) / 1.8)


cook_pause = Switch(
    name='Cook/Pause',
    icon='mdi:play-pause'
)

add_tuya_io(cook_pause, int, 1, 2)

power = Switch(
    name='Power',
    icon='mdi:power'
)

add_tuya_io(power, int, 1, 1)

delay_time = Number(
    name='Delay Time',
    number_range=range(0, 720),
    uom='minutes',
    icon='mdi:timer-pause'
)

add_tuya_io(delay_time, int, 2, 6)

status = Sensor(
    name='Status',
    icon='mdi:playlist-play',
    device_class='ENUM'
)

keep_warm_time = Number(
    name='Keep Warm Time',
    number_range=range(0, 120),
    uom='minutes',
    icon='mdi:timer-sync'
)

add_tuya_io(keep_warm_time, int, 2, 105)


@status.callback(trigger=tuya.received(4, 5))
def tasmota(value: int):
    return ['Ready', 'Delayed Cook', 'Cooking', 'Keep Warm', 'Off', 'Cooking Complete'][value]


DEVICE_CONFIG = {'run_config_post': f'def () return tasmota.cmd("TuyaSend0") end'}



from haco.control import Control
from haco.schema import Schema, Capability, AnnounceTopic, Tasmota
from haco.tools import get_range_pair

CELSIUS_C = 'C'
FAHRENHEIT_F = 'F'


class Climate(Control):
    NAME = 'climate'

    @classmethod
    def get_schema(cls):
        schema = Schema(
            capabilities=[
                Capability(
                    name='mode',
                    type=str
                ),
                Capability(
                    name='target_humidity'
                ),
                Capability(
                    name='temperature',
                    alias='target_temperature'
                ),
                Capability(
                    name='temperature_high'
                ),
                Capability(
                    name='temperature_low'
                ),
                Capability(
                    name='swing_mode',
                    type=str
                ),
                Capability(
                    name='preset_mode',
                    type=str
                ),
                Capability(
                    name='current_humidity',
                    ha=None,
                    tamota=Tasmota(announce_data=AnnounceTopic(key='{capability}_topic'))
                ),
                Capability(
                    name='current_temperature',
                    ha=None,
                    tamota=Tasmota(announce_data=AnnounceTopic(key='{capability}_topic'))
                ),
                Capability(
                    name='action',
                    type=str,
                    ha=None,
                    tamota=Tasmota(announce_data=AnnounceTopic(key='{capability}_topic'))
                )
            ]
        )
        return schema

    #    def init(name, entity_id, icon, temperature_unit, temperature_range, humidity_range, modes, preset_modes, fan_modes, swing_modes, precision, callbacks)

    def __init__(self, name=None, icon=None, temperature_unit=None, temperature_range=None, humidity_range=None,
                 modes=None, preset_modes=None, fan_modes=None, swing_modes=None, precision=None):
        self.temperature_unit = temperature_unit

        self.temperature_min, self.temperature_max = get_range_pair(temperature_range)
        self.humidity_min, self.humidity_max = get_range_pair(humidity_range)

        self.precision = precision
        # self.type = float if self.precision is None else int if self.precision == 1 else float if self.temperature_unit == 'C' else int

        self.fan_modes = fan_modes
        self.modes = modes
        self.preset_modes = preset_modes
        self.swing_modes = swing_modes

        super().__init__(name, icon=icon)

    @property
    def control_type(self):

        if self.precision == 1:
            return int

        if self.precision:
            return float

        unit = self.temperature_unit or CELSIUS_C

        if unit == CELSIUS_C:
            return float
        else:
            return int

    def get_config_ha_ex(self):
        data = {

            'fan_modes': self.fan_modes,
            'preset_modes': self.preset_modes,
            'swing_modes': self.swing_modes,
            'modes': self.modes,
            'min_temp': self.temperature_min,
            'max_temp': self.temperature_max,
            'min_humidity': self.humidity_min,
            'max_humidity': self.humidity_max,
            'precision': self.precision,
            'temperature_unit': self.temperature_unit

        }

        return data



import aiomqtt
import asyncio
import json
import os
from pathlib import Path

from haco import constants, tools
from haco.device import Device
from haco.load_configs import load_devices
from haco.tools import log_received

MQTT_PASSWORD = os.environ['MQTT_PASSWORD']
MQTT_HOST = os.environ['MQTT_HOST']
MQTT_PORT = int(os.environ['MQTT_PORT'])
MQTT_USERNAME = os.environ['MQTT_USERNAME']


async def listen(client, devices):
    TOPIC_ANNOUNCE = f"haco/{constants.BRANCH}/+/announce"
    await client.subscribe(TOPIC_ANNOUNCE)

    async with client.messages() as messages:

        async for message in messages:
            await log_received(message)
            if message.topic.matches(TOPIC_ANNOUNCE):

                if not message.payload:
                    msg = f'Topic {message.topic.value} got empty payload. It may have been deleted.'
                    tools.logger.info(msg)
                    continue

                data_announce = json.loads(message.payload)
                mac = data_announce['wifi']['mac']
                if mac not in devices:
                    msg = f'Got announce from device with no config: Hostname: {data_announce["hostname"]} MAC:{data_announce["mac"]}. Publishing config data.'
                    tools.logger.info(msg)
                    continue

                device: Device = devices[mac]

                if device.config_id and data_announce['config_id'] == device.config_id:
                    msg = f'Device {device.name} successfully configured. Hostname: {device.hostname} MAC:{device.mac}'
                    tools.logger.info(msg)
                    continue

                await device.bind(client, data_announce)
                devices[device.hostname] = device


            else:
                _, _, device_name, _, control, capability_id, platform, io = Path(message.topic.value).parts
                device = devices[device_name]
                await device.get(control).process_message(client, capability_id, platform, message)


async def start_mqtt():
    devices = load_devices()
    devices = {device.mac: device for device in devices}

    async with aiomqtt.Client(hostname=MQTT_HOST, port=MQTT_PORT, username=MQTT_USERNAME,
                              password=MQTT_PASSWORD) as client:
        await asyncio.gather(listen(client, devices))



def send(type_id, dp_id):
    return f'def (value,data) import string import uuid print(data["topic"], uuid.uuid4()) var cmd=string.format("TuyaSend{type_id} {dp_id},%s", value) return tasmota.cmd(cmd) end'


def cmd(mask):
    cmd = mask.format(value='%s')

    return f'def (value,data) import string var cmd=string.format("{cmd}", value) return tasmota.cmd(cmd) end'


def received(type_id, dp_id):
    return f'tuyareceived#dptype{type_id}id{dp_id}'



from haco import schema
from haco.control import Control


class Sensor(Control):
    NAME = 'sensor'

    @classmethod
    def get_schema(cls):
        return schema.Schema(
            capabilities=[
                schema.Capability(
                    name='state',
                    ha=None,
                    tamota=schema.Tasmota(announce_data=schema.AnnounceTopic(key='{io_ha}_topic'))
                )
            ]
        )

    def __init__(self, name, data_type=None, uom=None, icon=None, device_class=None):
        self.data_type = data_type
        self.device_class = device_class
        self.uom = uom

        super().__init__(name, icon=icon)

    @property
    def control_type(self):
        return self.data_type or (lambda v: v)

    def get_config_ha_ex(self):
        data = {
            'unit_of_measurement': self.uom,
            'device_class': self.device_class
        }

        return data



from importlib import reload

import importlib.util
import json
import os
from pathlib import Path

from haco import tools
from haco.control import Control
from haco.device import Device

CONFIGS_PATH = Path(os.environ['HACO_CONFIGS_PATH'])
OPTIONS_PATH = Path(os.environ['HACO_OPTIONS_PATH'])


def load_json(file_path):
    with open(file_path, 'r') as json_file:
        data = json.loads(json_file.read())
    return data


def load_controls_from_directory():
    if not CONFIGS_PATH.exists():
        tools.logger.warning(f'Configs path does not exist. Will be created: {CONFIGS_PATH}')
        CONFIGS_PATH.mkdir()

    options = load_json(OPTIONS_PATH)

    controls = []
    for data in options['assignments']:

        name = data['config']
        macs = data['identifier']

        if type(macs) is not list:
            macs = [macs]

        macs

        for mac in macs:

            args = {'mac': mac}

            py_file = CONFIGS_PATH / Path(name).with_suffix('.py')
            module_name = py_file.stem

            try:
                module = importlib.import_module(f'haco.configs.{module_name}')
                reload(module)
            except ImportError:
                spec = importlib.util.spec_from_file_location(module_name, str(py_file))
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)

            args['config_module'] = module

            for obj_name in dir(module):
                obj = getattr(module, obj_name)
                if obj_name == 'DEVICE_CONFIG':
                    args.update(obj)
                if isinstance(obj, Control):
                    args.setdefault('controls', [])
                    args['controls'].append(obj)

            controls.append(args)

    return controls


def load_devices():
    devices_args = load_controls_from_directory()
    devices = [Device(**args) for args in devices_args]
    return devices



from typing import List

from haco.control import Control


class Select(Control):
    NAME = 'select'

    def __init__(self, name, options: List[str], icon=None):
        self.options = options
        super().__init__(name, icon=icon)

    def get_config_ha_ex(self):
        data = {
            'options': self.options,
        }

        return data



import json
import logging
import typing
from pathlib import Path
from typing import List

from haco import tools, constants

DEVICE = 'development-eth'
CONTROL = 'DEVELOPMENT-DEHUMIDIFIER'

ON = 'ON'
OFF = 'OFF'


class ControlType:
    ...


class AnnounceDatum:
    DEFAULT_KEY = None
    DEFAULT_VALUE = None

    def __init__(self, key=None, value=None):
        self.platform = None
        self.key = key or self.DEFAULT_KEY
        self.value = value or self.DEFAULT_VALUE

    def get_config_ha(self):
        fields_data = dict(
            path=str(self.platform.capability.get_path()),
            # platform='ha',
            io=self.platform.IO_REVERSE,
            io_ha=self.platform.IO_HA,
            ha=HomeAssistant.PLATFORM,
            capability=self.platform.capability.name
        )

        filler = lambda s: s.format(**fields_data)
        data = {filler(self.key): filler(self.value)}
        return data


class AnnounceTopic(AnnounceDatum):
    DEFAULT_KEY = '{capability}_{io_ha}_topic'
    DEFAULT_VALUE = '{path}/{ha}/{io}'


# class AnnouncePayload(AnnounceDatum):
#
#     ON='ON'
#     OFF='OFF'
#     BOOL_MAP={True:ON,False:OFF}
#     DEFAULT_KEY = '{capability}_{io_ha}_topic'
#     DEFAULT_VALUE = '{path}/{ha}/{io}'
#
#     @classmethod
#     def from_bool(cls, value):
#         return cls.BOOL_MAP[value]

class Platform:
    IO = None
    PLATFORM = None
    IO_REVERSE = None
    PLATFORM_REVERSE = None

    IO_HA = None

    def __init__(self, announce_data: List[AnnounceDatum] | AnnounceDatum = None,
                 callback_default=tools.callback_default):
        self.capability: Capability = None
        if announce_data is None:
            announce_data = AnnounceTopic()
        self.announce_data = announce_data if type(announce_data) is list else [announce_data]
        for datum in self.announce_data:
            datum.platform = self

        self.callback_default = callback_default

        self.callback = None

    @property
    def callback_function_name(self):
        return f'{self.capability.alias}_{self.PLATFORM}'

    @property
    def capability_type(self):
        capability_type = self.capability.type

        if capability_type is ControlType:
            return self.capability.schema.control.control_type
        else:
            return capability_type

    def deserialize(self, value):
        return json.loads(value)

    def serialize(self, value):
        raise NotImplementedError()

    def get_handler_name(self):
        return f'callback_{self.capability.name}_{self.IO}'

    def handle(self, message):

        # handler = getattr(self.capability.schema.control, self.get_handler_name())

        payload_str = message.payload.decode('utf-8')
        payload = self.deserialize(payload_str)

        output_raw = self.get_callback_output(payload)

        output = self.serialize(self.convert_out(output_raw))

        return output

    def get_callback_output(self, data):
        raise NotImplementedError()

    def get_subscriptions(self):  # Map incoming to handler

        if not self.callback:
            return {}

        return {self.get_topic_subscription(): self.get_handler_name()}

    def get_config_ha(self):

        data = {}

        if not self.callback:
            return data

        for datum in self.announce_data:
            config = datum.get_config_ha()
            data.update(config)

        return data

    def get_config_tasmota(self):
        raise NotImplementedError()

    def convert_out(self, value):
        raise NotImplementedError()

    def convert_in(self, value):
        raise NotImplementedError()


class HomeAssistant(Platform):
    IO = 'in'
    PLATFORM = 'ha'

    IO_REVERSE = 'out'
    PLATFORM_REVERSE = 'tasmota'

    IO_HA = 'command'

    def get_callback_output(self, data):

        sigs = self.callback.get_function_data()
        return_type, return_exp = sigs.pop('return')

        if len(sigs) > 2:
            raise ValueError('Callback has too many args?')

        data = self.convert_in(data)
        data = [data]

        if 'control' in sigs:
            data.append(self.capability.schema.control)

        handler = self.callback.function
        output_raw = handler(*data)

        if return_type:
            output_raw = return_type(output_raw)

        return output_raw

    def get_topic_subscription(self):  # Map incoming to handler

        if not self.callback:
            return None

        return self.capability.get_path() / self.PLATFORM / self.IO_REVERSE

    def get_topic_publish(self):  # Map incoming to handler
        if not self.callback:
            return None
        return self.capability.get_path() / self.PLATFORM_REVERSE / self.IO

    def convert_out(self, value):
        return value

    def convert_in(self, value):
        return self.capability_type(value)

    def deserialize(self, value):
        return str(value)

    def serialize(self, value):
        return json.dumps(value)

    def get_config_tasmota(self):
        # Must be MQTT. Fixed trigger. Fixed topic.

        if not self.callback:
            return {}

        trigger = self.get_topic_publish()
        topic = trigger / 'output'

        config = {
            'topic': None,
            'trigger': str(trigger),
            'function': self.callback.function_tasmota,
            'type': 'MqttSubscription'
        }

        return config


class Tasmota(Platform):
    IO = 'out'
    PLATFORM = 'tasmota'

    IO_REVERSE = 'in'
    PLATFORM_REVERSE = 'ha'

    IO_HA = 'state'

    def get_callback_output(self, data):

        handler = self.callback.function

        sigs = self.callback.get_function_data()
        sig_return = sigs.pop('return')

        extra_args = set(data.keys()) - set(sigs.keys())

        for arg in extra_args:
            msg = f'Callback {self.callback.function_name} got argument ({arg}={extra_args[arg]}) not in callback function. Will be ignored.'
            logging.warning(msg)
            data.pop(arg)

        for name, (type_type, exp) in sigs.items():
            if type_type and type_type is not typing.Any:
                try:
                    data[name] = type_type(data[name])
                except TypeError as error:
                    logging.warning(f'Type conversion failed: {repr(error)}')

        if 'control' in sigs:
            data['control'] = self.capability.schema.control

        output_raw = handler(**data)
        return output_raw

    def get_topic_subscription(self):  # Map incoming to handler
        return self.capability.get_path() / self.PLATFORM / self.IO

    def get_topic_publish(self):  # Map incoming to handler
        return self.capability.get_path() / self.PLATFORM_REVERSE / self.IO_REVERSE

    def deserialize(self, value):
        return json.loads(value)

    def serialize(self, value):
        return str(value)

    def convert_out(self, value):
        return self.capability_type(value)

    def convert_in(self, value):
        return value

    def get_config_tasmota(self):
        # Any type. Any trigger. Fixed topic.

        if not self.callback:
            return {}

        config = {
            'topic': str(self.get_topic_subscription()),
            'trigger': self.callback.trigger,
            'function': self.callback.function_tasmota,
            'type': self.callback.type_id
        }

        return config


class Default:
    ...


class Capability:

    def __init__(self, name, type=ControlType, tamota: Tasmota = Default, ha: HomeAssistant = Default, alias=None):
        self.name = name
        self.alias = alias or name
        self.type = type
        self.schema = None

        self.tasmota = tamota if tamota is not Default else Tasmota()
        self.ha = ha if ha is not Default else HomeAssistant()
        self.platforms = [io for io in {self.tasmota, self.ha} if io]

        for platform in self.platforms:
            platform.capability = self

        self.platforms_dict = {io.PLATFORM: io for io in self.platforms}
        self.get = self.platforms_dict.get

    def get_callback_function_names(self):
        return {platform.callback_function_name for platform in self.platforms}

    def get_path(self):
        return self.schema.get_path() / self.name

    def get_config_ha(self):
        data = {}
        for io in self.platforms:
            config = io.get_config_ha()
            data.update(config)
        return data

    def get_config_tasmota(self):
        return [config for io in self.platforms if (config := io.get_config_tasmota())]

    def get_subscriptions(self):
        data = {}
        for io in self.platforms:
            data.update(io.get_subscriptions())
        return data


class Schema:

    def __init__(self, capabilities: List[Capability]):
        self.capabilities = capabilities

        self.capabilities_dict = {capability.name: capability for capability in self.capabilities}
        self.capabilities_dict_alias = {capability.alias: capability for capability in self.capabilities}
        self.get = self.capabilities_dict.get
        self.get_alias = self.capabilities_dict_alias.get

        for capability in self.capabilities:
            capability.schema = self

        self.control = None

    def get_callback_function_names(self):
        names = set()
        for capability in self.capabilities:
            names.update(capability.get_callback_function_names())
        return names

    def get_path(self):
        return Path(f"haco") / constants.BRANCH / self.control.device.hostname / 'control' / self.control.name_sanitized

    def get_config_ha(self):
        data = {}
        for capability in self.capabilities:
            data.update(capability.get_config_ha())
        return data

    def get_config_tasmota(self):
        config = []
        for capability in self.capabilities:
            config += capability.get_config_tasmota()
        return config

    def get_subscriptions(self):
        data = {}
        for capability in self.capabilities:
            data.update(capability.get_subscriptions())
        return data
